"""BFGS method for extra task 2i"""
import timeit

import numpy as np
import matplotlib.pyplot as plt
from casadi import SX, Function, jacobian

from Phi import Phi
from finite_difference import finite_difference
from i_trick import i_trick
from Phi_FAD import Phi_FAD
from Phi_BAD import Phi_BAD
from Phi_casadi import derive_Phi_casadi


# GRADIENT = "forward_AD"
GRADIENT = "backward_AD"
# GRADIENT = "i_trick"
# GRADIENT = "finite_differences"
# GRADIENT = "casadi"

param = {
    'N' : 200,     # number of discretization steps
    'x0': 2,       # initial condition on state
    'T' : 5,       # terminal time
    'q' : 50,      # weight on terminal state
}
h = param['T']/param['N']


def obj(U):
    return U.T@U + Phi(U, param)

U = np.zeros((param['N'],1))   # initial controls
B = np.eye(param['N'])         # initial Hessian

# Printing header: iterate number, gradient norm, objective value, step norm, stepsize
print("It.\t | ||grad_f||| f\t\t | ||dvar||\t | t  \n")

# Get the objective value and the gradient
if GRADIENT == "finite_differences":
    [f, J] = finite_difference(Phi, U, param)
elif GRADIENT == "i_trick":
    [f, J] = i_trick(Phi, U, param)
elif GRADIENT == "backward_AD":
    [f, J] = Phi_BAD(U, param)
elif GRADIENT == "forward_AD":
    [f, J] = Phi_FAD(U, param)
elif GRADIENT == "casadi":
    Uvar  = SX.sym("u", param['N'])
    phi_expr = param['x0']
    for i in range(param['N']):
        phi_expr = phi_expr + h*((1 - phi_expr)*phi_expr + Uvar[i])

    phi_expr = param['q'] * phi_expr**2
    Phi_func = Function("Phi", [Uvar], [phi_expr])
    J_func   = Function("J", [Uvar], [jacobian(phi_expr,Uvar)])
    f = Phi_func(U).full()
    J = J_func(U).full()


J = J + 2*U.T
f = f + np.sum(U**2)

maxit = 300    # max trials of iteration
tol   = 1e-3   # tolerance
tlog  = 0      # execution time


# Iterative loop
for k in range(maxit):
    # Obtaining search direction
    dx = -np.linalg.solve(B, J.T)

    # We check Armijo"s condition to find a step length
    t     = 1.0     # Initial step length
    beta  = 0.8     # Shrinking factor
    gamma = 0.1     # Minimal decrease requirement


    U_new = U + t*dx  # Candidate for the next step
    # Iterate on Armijo"s condition
    while np.any(obj(U_new) > f + gamma * t * J@dx):
        t = beta * t
        U_new = U + t*dx
    #var_new = var + dvar
    
    time_ref = timeit.default_timer()
    if GRADIENT == "finite_differences":
        [f_new, J_new] = finite_difference(Phi , U_new, param)
    elif GRADIENT == "i_trick":
        [f_new, J_new] = i_trick(Phi , U_new, param)
    elif GRADIENT == "backward_AD":
        [f_new, J_new] = Phi_BAD(U_new, param)
    elif GRADIENT == "forward_AD":
        [f_new, J_new] =  Phi_FAD(U_new, param)
    elif GRADIENT == "casadi":
        f_new = Phi_func(U_new).full()
        J_new = J_func(U_new).full()
    tlog = tlog + timeit.default_timer() - time_ref
    
    J_new = J_new + 2*U_new.T
    f_new = f_new + np.sum(U_new**2)

    # Updating Hessian according to BFGS formula
    s = U_new - U
    z = J_new.T - J.T
    B = B - (B @ (s @ s.T) @ B) / (s.T @ B @ s) + (z @ z.T) / (s.T @ z)

    # Updating variables
    U = U_new
    J = J_new
    f = f_new

    # Every 10 iterations print the header again
    if k % 10 == 0:
        print("\n")
        print("It.\t | ||grad_f||| f\t\t | ||dvar||\t | t  \n")
    
    # Print some useful information
    print(r"%d\t | %8.5f\t | %8.5f\t | %8.5f\t | %f \n" % (k, np.linalg.norm(J), f, np.linalg.norm(dx), t))
    
    if np.linalg.norm(J) < tol:
        print("Convergence achieved.")
        break


t = np.linspace(0, param['T'], param['N']+1)

# plot controls
plt.step(t, np.vstack((U,U[-1])), "r")

# reconstruct states
X = np.zeros((param['N']+1,1))
X[0] = param['x0']

for i in range(param['N']):
    X[i+1] = X[i] + h*((1-X[i])*X[i] + U[i])

# plot trajectory
plt.plot(t, X)

plt.legend(["control","state"])
plt.title("Optimal solution")
plt.xlabel("Time")

print("")
print("TOTAL TIME SPENT IN DERIVATIVES")
print("")
print(tlog)


plt.show()
